// for finding memory leaks in debug mode with Visual Studio 
#if defined _DEBUG && defined _MSC_VER
#include <crtdbg.h>
#endif

#include <stdint.h>
#include <stdbool.h>
#include <math.h> // modf()
#ifndef _WIN32
#include <unistd.h> // usleep()
#endif
#include "pt_header.h"
#include "pt_helpers.h"
#include "pt_visuals.h"
#include "pt_scopes.h"
#include "pt_sampler.h"
#include "pt_palette.h"
#include "pt_tables.h"

// this uses code that is not entirely thread safe, but I have never had any issues so far...

static volatile bool scopesReading;
static uint32_t scopeTimeLen, scopeTimeLenFrac;
static uint64_t timeNext64, timeNext64Frac;
static SDL_Thread *scopeThread;

scopeChannel_t scope[4]; // global

extern bool forceMixerOff;    // pt_audio.c
extern uint32_t *pixelBuffer; // pt_main.c

int32_t getSampleReadPos(uint8_t ch, uint8_t smpNum)
{
	int32_t pos;
	scopeChannel_t *sc;
	moduleSample_t *s;
	
	sc = &scope[ch];

	if (sc->retriggered)
		return (-1); // scope is latching at the moment, try again later

	pos = sc->pos; // cache it

	if (sc->active && (pos >= 2))
	{
		s = &modEntry->samples[smpNum];

		/* get real sampling position regardless of where the scope data points to
		** sc->data changes during loop, offset and so on, so this has to be done (sadly, because it's really hackish)
		*/

		pos = (int32_t)(&sc->data[pos] - &modEntry->sampleData[s->offset]);
		if (pos >= s->length)
			return (-1);

		// do this check again here - check if scope is latching at the moment
		if (sc->retriggered)
			return (-1);

		return (pos);
	}

	return (-1);
}

void updateScopes(void)
{
	uint8_t i;
	int32_t pos, len;
	uint32_t period;
	scopeChannel_t *sc;

	if (editor.isWAVRendering)
		return;

	for (i = 0; i < AMIGA_VOICES; i++)
	{
		sc = &scope[i];
		if (sc->retriggered)
		{
			sc->pos     = 0;
			sc->posFrac = 0;

			// data/length is already set from replayer thread (important)
			sc->loopFlag  = sc->newLoopFlag;
			sc->loopStart = sc->newLoopStart;

			sc->didSwapData = false;
			sc->active      = true;

			sc->retriggered = false;
		}

		if (sc->active)
		{
			period = (uint32_t)(modEntry->channels[i].n_period);
			if (period != 0)
			{
				if (period < 113)
					period = 113;

				sc->posFrac += (((3546895 / period) << 16) / SCOPE_HZ);
			}

			if (sc->posFrac >= 65536)
			{
				pos = sc->pos + (sc->posFrac >> 16);
				sc->posFrac &= 0xFFFF;

				if (pos >= sc->length)
				{
					// sample reached end, simulate Paula register update (sample swapping)

					/* wrap phase around one time with current length, then set new length
					** and wrap around it (handles one-shot loops and sample swapping)
					*/
					len  = sc->length;
					pos -= len;
					len  = sc->newLength;

					if (len > 0)
						pos %= len;

					sc->length      = len;
					sc->pos         = pos;
					sc->data        = sc->newData;
					sc->loopFlag    = sc->newLoopFlag;
					sc->loopStart   = sc->newLoopStart;
					sc->didSwapData = true;
				}
				else
				{
					// no sample end, update position only
					sc->pos = pos;
				}
			}
		}
	}
}

void drawScopes(void)
{
	uint8_t y;
	int16_t scopeData;
	int32_t i, x, readPos;
	uint32_t *dstPtr, *scopePtr, scopePixel;
	scopeChannel_t cachedScope, *sc;

	scopesReading = true;
	if (editor.ui.visualizerMode == VISUAL_QUADRASCOPE)
	{
		// --- QUADRASCOPE ---

		scopePtr = &pixelBuffer[(71 * SCREEN_W) + 128];
		for (i = 0; i < AMIGA_VOICES; ++i)
		{
			// cache scope channel to lower thread race condition issues
			memcpy(&cachedScope, &scope[i], sizeof (scopeChannel_t));
			sc = &cachedScope;

			// render scope
			if (sc->active && ((sc->length > 2) || sc->loopFlag) &&
				(sc->volume != 0) && !editor.muted[i] && (sc->data != NULL))
			{
				// scope is active

				scope[i].emptyScopeDrawn = false;

				// draw scope background

				dstPtr = &pixelBuffer[(55 * SCREEN_W) + (128 + (i * (SCOPE_WIDTH + 8)))];
				scopePixel = palette[PAL_BACKGRD]; // this palette can change

				for (y = 0; y < SCOPE_HEIGHT; ++y)
				{
					for (x = 0; x < SCOPE_WIDTH; ++x)
						dstPtr[x] = scopePixel;

					dstPtr += SCREEN_W;
				}

				// render scope data

				scopePixel = palette[PAL_QADSCP];

				readPos = sc->pos;
				if (sc->loopFlag)
				{
					// loop enabled

					for (x = 0; x < SCOPE_WIDTH; ++x)
					{
						if (sc->didSwapData)
						{
							if (sc->length > 0)
								readPos %= sc->length; // sc->data = loopStartPtr, wrap readPos to 0
						}
						else if (readPos >= sc->length)
						{
							readPos = sc->loopStart; // sc->data = sampleStartPtr, wrap readPos to loop start
						}

						scopeData = (sc->data[readPos++] * sc->volume) >> 8;
						scopePtr[(scopeData * SCREEN_W) + x] = scopePixel;
					}
				}
				else
				{
					// no loop

					for (x = 0; x < SCOPE_WIDTH; ++x)
					{
						if (readPos >= sc->length)
						{
							scopePtr[x] = scopePixel; // end of data, draw center pixel
						}
						else
						{
							scopeData = (sc->data[readPos++] * sc->volume) >> 8;
							scopePtr[(scopeData * SCREEN_W) + x] = scopePixel;
						}
					}
				}
			}
			else
			{
				// scope is inactive, draw empty scope once until it gets active again

				sc = &scope[i];
				if (!sc->emptyScopeDrawn)
				{
					// draw scope background

					dstPtr     = &pixelBuffer[(55 * SCREEN_W) + (128 + (i * (SCOPE_WIDTH + 8)))];
					scopePixel = palette[PAL_BACKGRD];

					for (y = 0; y < SCOPE_HEIGHT; ++y)
					{
						for (x = 0; x < SCOPE_WIDTH; ++x)
							dstPtr[x] = scopePixel;

						dstPtr += SCREEN_W;
					}

					// draw line

					scopePixel = palette[PAL_QADSCP];
					for (x = 0; x < SCOPE_WIDTH; ++x)
						scopePtr[x] = scopePixel;

					sc->emptyScopeDrawn = true;
				}
			}

			scopePtr += (SCOPE_WIDTH + 8);
		}
	}
	scopesReading = false;
}

static int32_t SDLCALL scopeThreadFunc(void *ptr)
{
	int32_t time32;
	uint32_t diff32;
	uint64_t time64;
	double dTime;

	// this is needed for scope stability (confirmed)
	SDL_SetThreadPriority(SDL_THREAD_PRIORITY_HIGH);

	// set next frame time
	timeNext64     = SDL_GetPerformanceCounter() + scopeTimeLen;
	timeNext64Frac = scopeTimeLenFrac;

	while (editor.programRunning)
	{
		updateScopes();

		time64 = SDL_GetPerformanceCounter();
		if (time64 < timeNext64)
		{
			assert((timeNext64 - time64) <= 0xFFFFFFFFULL);
			diff32 = (uint32_t)(timeNext64 - time64);

			// convert to microseconds and round to integer
			dTime = diff32 * editor.dPerfFreqMulMicro;
			double2int32_round(time32, dTime);

			// delay until we have reached next tick
			if (time32 > 0)
				usleep(time32);
		}

		// update next tick time
		timeNext64 += scopeTimeLen;

		timeNext64Frac += scopeTimeLenFrac;
		if (timeNext64Frac >= (1ULL << 32))
		{
			timeNext64++;
			timeNext64Frac &= 0xFFFFFFFF;
		}
	}

	(void)(ptr); // make compiler happy

	return (true);
}

bool initScopes(void)
{
	double dInt, dFrac;

	// calculate scope time for performance counters and split into int/frac
	dFrac = modf(editor.dPerfFreq / SCOPE_HZ, &dInt);

	// integer part
	scopeTimeLen = (uint32_t)(dInt);

	// fractional part scaled to 0..2^32-1
	scopeTimeLenFrac = (uint32_t)(MIN(round((UINT32_MAX + 1.0) * dFrac), UINT32_MAX));

	scopeThread = SDL_CreateThread(scopeThreadFunc, NULL, NULL);
	if (scopeThread == NULL)
	{
		showErrorMsgBox("Couldn't create scope thread!");
		return (false);
	}

	SDL_DetachThread(scopeThread);
	return (true);
}

void waitOnScopes(void)
{
	while (scopesReading);
}

void clearScopes(void)
{
	uint8_t i;

	waitOnScopes();
	memset(scope, 0, sizeof (scope));

	for (i = 0; i < AMIGA_VOICES; ++i)
		scope[i].length = scope[i].newLength = 2;
}
